#!/usr/bin/env python
#
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2020 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a Python plug-in for GIMP. It is a Python implementation of the amazing Bevel Reflect Logo script (version 1.6) by GnuTux (2013) in order to apply the MLBevelLogo2 plug-in to all the character layers in a custom font file.
# It simply applies the work principally of GnuTux (2013) and MareroQ (2020) in batch mode.
#
# It can be executed by selecting the menu option: 'Filters>>Custom Font Tools>>Custom Font Bevel Reflect Version 1.03...

# version 1.00
#     The filter will complain if it cannot locate MLBevelLogo2 and end
#     Most of the options available from the user dialog mirror the options available in the original script dialog by GnuTux
#     with the following additional options:
#        Option to trim pixels from the character should the user need to tidy up the character edges
#        Option to desaturate output from BRL to apply the generated highlights to an existing custom font
#        Option to generate different emaps for different character groups
#        Option to save any generated emaps as a new image

# version 1.01
#		Fixed trim option - was trimming the wrong layer
# version 1.02
#        Blind attempt to resolve calling error on SamJ portable
# version 1.03
#        Resolved calling error on SamJ portable



from gimpfu import *
#from array import array
import math
import random
import time


# Some global colours we may need to define

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)



def do_stuff(image, layer, trim_value, desaturate_option, slope_options, 
            flat_bevel_smoothing, flat_bevel_depth, round1_bevel_depth, round2_bevel_depth, 
            diffuse_shiny, contour, blur_depth, edge_smoothing, emap_name, map_type, 
            paint_mode, blend_direction, blend_shape, light_direction, specular_highlight, 
            detail_level, horizontal_texture_size, vertical_texture_size, blur_radius, 
            emap_to_character, keep_own_layer):

    #### Main Filter Proc
	
    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)
	
    # define strings to hold characters for grouping options
    upper_case = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    lower_case = "abcdefghijklmnopqrstuvwxyz"
    number_case = "0123456789$" + unichr(8364) + unichr(188) + unichr(189) + unichr(190) + unichr(162) # Euro sign, one quarter sign, one half sign, three quarters sign, cent sign
    number_case += unichr(176) + unichr(185) + unichr(178) + unichr(179) + unichr(183) + unichr(215) + unichr(247) # degree sign, superscript 1, superscript 2, superscript 3, middle dot, multiplication, division
    number_case += unichr(163) # pound sterling
    punctuation_case = ".,:;'!?\"" + unichr(161) + unichr(191)  + unichr(8220) + unichr(8216) #  inverted exclamation + inverted question + smart right hand double quote + smart right hand single quote
	
    vowel_case, consonant_case = "AEIOUaeiou","BCDFGHJKLMNPQRSTVWXYZbcdfghjklmnpqrstvwxyz"
	
    french_uc_vowels = unichr(192) + unichr(194) + unichr(196) + unichr(198) # A grave + A circumflex + A diaeresis + AE ligature
    french_uc_vowels += unichr(200) + unichr(201) + unichr(202) + unichr(203) + unichr(206) + unichr(207) # E grave + E acute + E circumflex + E diaeresis + I circumflex + I diaeresis
    french_uc_vowels += unichr(212) + unichr(217) + unichr(219) + unichr(220) # O circumflex + U grave + U circumflex + U diaeresis
    french_lc_vowels = unichr(224) + unichr(226) + unichr(228) + unichr(230) # a grave + a circumflex + a diaeresis + ae ligature
    french_lc_vowels += unichr(232) + unichr(233) + unichr(234) + unichr(235) + unichr(238) + unichr(239) # e grave + e acute + e circumflex + e diaeresis + i circumflex + i diaeresis
    french_lc_vowels += unichr(244) + unichr(249) + unichr(251) + unichr(252) + unichr(338) + unichr(339) + unichr(376) + unichr(255) # o circumflex + u grave + u circumflex + u diaeresis + OE ligature + oe ligature + Y dieresis + y dieresis

    french_uc_cons = unichr(199) # C cedilla
    french_lc_cons = unichr(231) # c cedilla
	
    german_uc_vowels = unichr(196) + unichr(214) + unichr(220) # A diaeresis + O diaeresis + U diaeresis
    german_lc_vowels = unichr(228) + unichr(246) + unichr(252) # a diaeresis + o diaeresis + u diaeresis
	
    german_uc_cons =  unichr(223) # Eszett
	
    italian_uc_vowels = unichr(192) + unichr(200) + unichr(201) + unichr(204) + unichr(210) + unichr(217) #  A grave + E grave + E acute + I grave + O grave + U grave
    italian_lc_vowels = unichr(224) + unichr(232) + unichr(233) + unichr(236) + unichr(242) + unichr(249) # a grave + e grave + e acute + i grave + o grave + u grave
	
    polish_uc_vowels = unichr(260) + unichr(280) + unichr(211) + unichr(218) + unichr(220) + unichr(225)  # A ogonek + E ogonek + O acute
    polish_lc_vowels = unichr(261) + unichr(263) + unichr(324) + unichr(243)# a ogonek + e ogonek + o acute
	
    polish_uc_cons = unichr(262) + unichr(323) + unichr(346) + unichr(377) + unichr(379) # C acute + L stroke + N acute + S acute + Z acute + Z dot above
    polish_lc_cons = unichr(321) + unichr(281) + unichr(322) + unichr(347) + unichr(378) + unichr(380)  # c acute + l stroke + n acute + s acute + z acute + z dot above
	
    spanish_uc_vowels =  unichr(193) + unichr(201) + unichr(205) + unichr(211) + unichr(218) + unichr(220) # A acute + E acute + I acute + O acute + U Acute + U diaeresis
    spanish_lc_vowels = unichr(233) + unichr(237) + unichr(241) # a acute + e acute + i acute 

    spanish_uc_cons = unichr(209) # N tilde
    spanish_lc_cons = unichr(241) # n tilde
	
	
    upper_case += french_uc_vowels + french_uc_cons + german_uc_vowels + german_uc_cons + italian_uc_vowels + polish_uc_vowels + polish_uc_cons
    lower_case += french_lc_vowels + french_lc_cons + german_lc_vowels + italian_lc_vowels + polish_lc_vowels + polish_lc_cons
	
    vowel_case += french_uc_vowels + french_lc_vowels + german_uc_vowels + german_lc_vowels + italian_uc_vowels + italian_lc_vowels + polish_uc_vowels + polish_lc_vowels
    consonant_case += french_uc_cons + french_lc_cons + german_uc_cons + polish_uc_cons + polish_lc_cons
	
	
    # check for MLBevelReflect2
    plugin = False
    num_plugins, menu_path, num_plugins, plugin_accelerator, num_plugins, plugin_location, num_plugins, plugin_imgtype, num_plugins, plugin_time, num_plugins, plugin_real_name = pdb.gimp_plugins_query("reflect")
    
    if "bevelreflect2" in str(plugin_real_name): plugin = True
    
    original_active_layer = image.active_layer # get active layer
	
	# set variables
    bevel_gradient = pdb.gimp_context_get_gradient()
    im_height = pdb.gimp_image_height(image)
    emap_final_size = int(im_height/2)
    auto_stretch_HSV = True
    turbulent_noise = True
    keep_layers = False
    # get random number to use as turbulence seed
    random.seed()
    random.seed()
    random_number = int(time.time() )
    
    own_emap = False # flag for generating emap/s
    if not (emap_name.endswith(".xcf") or emap_name.endswith(".XCF") or emap_name.endswith(".png") or emap_name.endswith(".PNG")):
        emap_name = "none"
        emap_drawable = None
    if emap_name == "": emap_name = "none"

    if emap_name=="none":
        # Prepare for generating a new emap
        own_emap = True # set flag
        # save current colours
        saved_fg_color = pdb.gimp_context_get_foreground()
        saved_bg_color = pdb.gimp_context_get_background()
	
        # get current active layer
        layer_id = pdb.gimp_image_get_active_layer(image)
        active_layer_id = pdb.gimp_image_get_active_layer(image)
	
        # Get Parent of active layer
        parent = pdb.gimp_item_get_parent(active_layer_id)
        # Save Active Layer Position
        act_lay_pos = pdb.gimp_image_get_item_position(image, active_layer_id)
        # calculate size of required emap
        im_height = pdb.gimp_image_height(image)
        emap_final_size = int(im_height)

        e_loop_max = 1
        if emap_to_character>0: e_loop_max = 5
        eloop = 0
        while eloop < e_loop_max:
            # get random number to use as turbulence seed
            random.seed()
            random.seed()
            random_number = int(time.time() )
            emap_drawable = proc_generate_emap(image, random_number, active_layer_id, emap_drawable, emap_final_size, parent, act_lay_pos, 
                    detail_level, horizontal_texture_size, vertical_texture_size, map_type, 
                    bevel_gradient, saved_fg_color, saved_bg_color, blend_direction, specular_highlight, 
                    light_direction, paint_mode, blend_shape, blur_radius) 
            if eloop == 0:
                emap_list = [emap_drawable]
            else:
                emap_list.append(emap_drawable)
            eloop += 1
    else:
        # load emap
        emap_img = pdb.gimp_file_load(emap_name, emap_name)
        num_loaded_layers, loaded_layers = pdb.gimp_image_get_layers(emap_img)
        loaded_layer_ids = emap_img.layers
        eloop = 0
        if num_loaded_layers>5: 
            e_loop_max = 5
        else:
		    e_loop_max = num_loaded_layers
			
        while eloop < e_loop_max:
            emap_drawable = loaded_layer_ids[eloop]
            if eloop == 0:
                emap_list = [emap_drawable]
            else:
                emap_list.append(emap_drawable)
            eloop += 1
		
        emap_drawable = emap_list[0]
	
    # get layer info
    loop_ptr = 1
    first_char_layer_name = "none"
    num_layers, layer_ids = pdb.gimp_image_get_layers(image)
    loop = num_layers - loop_ptr
    layer_names_list = image.layers

	
    visi_loop, first_char_layer  = 0, -1
    other_layers,num_char_layers = 0,0
    apply_to_custom = False
	
	
    while visi_loop < num_layers:
	    # get info about layer we are working on
        layer_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
        layer_name = layer_name.decode('utf-8')
		
		
        test_for_custom_layer = fn_test_for_custom_layer(image, layer_names_list[visi_loop], layer_name)
			
        if test_for_custom_layer:
		
            apply_to_custom = True
			
            num_char_layers += 1
            if first_char_layer < 0: 
                first_char_layer = visi_loop
                first_char_layer_name = layer_name
                layer_names_list[visi_loop].set_offsets(0, 0)
		
        else:
                
            face_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            if other_layers == 0: other_layer_list = [face_name]
		
            other_layers += 1
			
        pdb.gimp_item_set_visible(layer_names_list[visi_loop],False) # turn off visibility
        visi_loop += 1
			
			
			
    if not apply_to_custom and plugin == True:	
        message_text = "No custom font character layers."
        old_handler = pdb.gimp_message_get_handler()	
        pdb.gimp_message_set_handler(MESSAGE_BOX)
        pdb.gimp_message(message_text)
        pdb.gimp_message_set_handler(old_handler)
		
    if not plugin:
        message_text = "BevelReflect2 plugin cannot be found.\nUnable to continue."
        apply_to_custom = False
        old_handler = pdb.gimp_message_get_handler()	
        pdb.gimp_message_set_handler(MESSAGE_BOX)
        pdb.gimp_message(message_text)
        pdb.gimp_message_set_handler(old_handler)
		
			
    if apply_to_custom:
        active_to_new = False		
				
    while loop >= 0 and apply_to_custom:
	    # get info about layer we are working on for custom font mode
        layer_id = layer_names_list[loop]
        layer_name = pdb.gimp_item_get_name(layer_names_list[loop])
		
        #test_for_custom_layer = fn_test_for_custom_layer(image, layer_names_list[loop], layer_name)
	
        # cater for unicode characters above normal range
        display_layer_name =  pdb.gimp_item_get_name(layer_names_list[loop])
        display_layer_name = display_layer_name.decode('utf-8')	
	
        # check we have the layer id of the currently active layer
        pdb.gimp_image_set_active_layer(image, layer_id)
        layer_id = pdb.gimp_image_get_active_layer(image)
        active_layer_id = pdb.gimp_image_get_active_layer(image)			
				
        test_for_custom_layer = fn_test_for_custom_layer(image, layer_id, layer_name)
			
        if test_for_custom_layer:
            # ensure we have the active layer and make it visible
            active_layer_id = pdb.gimp_image_get_active_layer(image)
            pdb.gimp_item_set_visible(active_layer_id, True)
			
            if emap_to_character == 2 and len(emap_list)<5: # check that there are enough emaps for this feature
                emap_to_character = 0 # reset to first drawable
                
            if emap_to_character == 3 and len(emap_list)< 2: # check that there are enough emaps for this feature
                emap_to_character = 0 # reset to first drawable
			
            if emap_to_character == 0: # use only one drawable for all characters
                emap_drawable = emap_list[emap_to_character]
				
            elif emap_to_character == 1: # use random emap
                emap_drawable = emap_list[0] # fall-back choice of drawable
                emap_list_len = len(emap_list)
                if emap_list_len>1:
                    #if layer_name[0] in 
                    random.seed()
                    emap_drawable = random.choice(emap_list)
					
            elif emap_to_character == 2: # use character groupings
                if display_layer_name[0] in upper_case: # use first emap for upper-case 
                    emap_drawable = emap_list[0]
                elif display_layer_name[0] in lower_case: # use second emap for lower-case 
                    emap_drawable = emap_list[1]
                elif display_layer_name[0] in number_case: # use third emap for number characters
                    emap_drawable = emap_list[2]
                elif display_layer_name[0] in punctuation_case: # use fourth emap for punctuation characters
                    emap_drawable = emap_list[3]
                else:
                    emap_drawable = emap_list[4] # use fifth emap for all other characters
            elif emap_to_character == 3: # use vowel/consonant groupings
                if display_layer_name[0] in vowel_case: # use first emap for vowels 
                    emap_drawable = emap_list[0]
                elif display_layer_name[0] in consonant_case: # use second emap for consonants 
                    emap_drawable = emap_list[1]
                else:
                    emap_drawable = emap_list[2] # use fifth emap for all other characters
                    

            active_layer_id = proc_bevel_reflect(0,image, active_layer_id, slope_options, 
                                                flat_bevel_smoothing, flat_bevel_depth, round1_bevel_depth, round2_bevel_depth, 
                                                diffuse_shiny, contour, blur_depth, edge_smoothing, emap_name, map_type, 
                                                paint_mode, blend_direction, blend_shape, bevel_gradient, 
                                                light_direction, specular_highlight, detail_level, horizontal_texture_size, vertical_texture_size, 
                                                emap_final_size, blur_radius, auto_stretch_HSV, turbulent_noise, 
                                                keep_layers, trim_value, desaturate_option, random_number, emap_drawable)
				
				
        if (num_layers - loop_ptr) > 0: pdb.gimp_item_set_visible(active_layer_id, False)
			
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        pdb.gimp_image_resize_to_layers(image)
			
        loop_ptr += 1
        loop = num_layers - loop_ptr
		
    #keep_own_layer = False
    if own_emap: 
        if keep_own_layer:
            pdb.gimp_selection_none(image) # Clear selection
            success = pdb.gimp_edit_copy(emap_list[0])
            if success: 
                new_image = pdb.gimp_edit_paste_as_new_image()
                if len(emap_list)>1:
                    eloop = 1
                    e_loop_maxloop = len(emap_list)
                    while eloop<len(emap_list):
                        layer_copy = pdb.gimp_layer_new_from_drawable(emap_list[eloop],new_image)
                        pdb.gimp_image_insert_layer(new_image, layer_copy, None, -1)
                        
                        eloop += 1
				
                new_display = pdb.gimp_display_new(new_image)
        eloop = 0
        while eloop<len(emap_list):
            
            pdb.gimp_image_remove_layer(image, emap_list[eloop]) # and remove own emap layer/s
            eloop+=1
		
    pdb.gimp_image_resize_to_layers(image)
	
    if first_char_layer_name <> "none": # the character layer to make visible on completion
        first_char_layer_id = pdb.gimp_image_get_layer_by_name(image, first_char_layer_name)
        pdb.gimp_item_set_visible(first_char_layer_id, True)
        pdb.gimp_image_set_active_layer(image,first_char_layer_id)
    else:	
        active_layer = image.active_layer
        pdb.gimp_item_set_visible(original_active_layer, True)
        pdb.gimp_image_set_active_layer(image,original_active_layer)
	
	
	
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()	
	
######################################################################################################################################

def proc_apply_curves(layer, contour):
    # apply curves 
    contour = contour - 1
    # number of x,y coordinates in spline for each contour style
    curve_num_segments = [6, 6, 6, 18, 8, 8, 8, 8, 18, 16, 14, 14]
    # list containing actual x,y coordinates
    curve_xy = [0, 0, 127, 255, 255, 0] + [0, 0, 02, 255, 255, 0] + [0, 0, 250, 255, 255, 0] + [0, 0, 33, 8, 64, 38, 97, 102, 128, 166, 158, 209, 191, 235, 222, 247, 255, 255] + [0, 0, 16, 255, 32, 0, 125, 255] + [0, 0, 64, 255, 127, 0, 255, 255] + [0, 0, 95, 255, 250, 0, 255, 255] + [0, 0, 95, 255, 253, 0, 255, 255] + [0, 0, 33, 110, 64, 237, 97, 240, 128, 138, 158, 33, 191, 5, 222, 99, 255, 255] + [0, 0, 42, 255, 43, 0, 80, 32, 151, 255, 164, 0, 223, 60, 255, 255] + [0, 0, 26, 255, 27, 0, 79, 255, 108, 0, 160, 255, 255, 255] + [0, 0, 26, 255, 27, 0, 152, 255, 162, 0, 237, 59, 255, 255]
    # pointer to appropriate contour with curve_xy list
    points_ptr = [0, 6, 12, 18, 36, 42, 50, 58, 66, 84, 100, 114]
    # replaces deprecated curves-spline pdb call in original script
    pdb.gimp_drawable_curves_spline(layer, 4, curve_num_segments[contour], curve_xy[points_ptr[contour]])
    

######################################################################################################################################
def proc_bevel_reflect(run_mode, image, active_layer_id, slope_options, 
                        flat_bevel_smoothing, flat_bevel_depth, round1_bevel_depth, round2_bevel_depth, 
                        diffuse_shiny, contour, blur_depth, edge_smoothing, emap_name, map_type, 
                        paint_mode, blend_direction, blend_shape, bevel_gradient, 
                        light_direction, specular_highlight, detail_level, horizontal_texture_size, vertical_texture_size, 
                        emap_final_size, blur_radius, auto_stretch_HSV, turbulent_noise, 
                        keep_layers, trim_value, desaturate_option, random_number, emap_drawable):

    #### The following is basically GnuTux's code in Python
    #### active_layer_id is the logo-layer in the original script

    # get character layer name for later rename
    character_name = pdb.gimp_item_get_name(active_layer_id)
	
    saved_fg_color = pdb.gimp_context_get_foreground()
    saved_bg_color = pdb.gimp_context_get_background()
	
    # Get Parent of current character drawable
    parent = pdb.gimp_item_get_parent(active_layer_id)

    # Selection Check - if a user selection exists save it to later restore
    saved_flag = False
    sel_flag = pdb.gimp_selection_is_empty(image)
    if sel_flag == False:
        saved_selection = pdb.gimp_selection_save(image)
        pdb.gimp_selection_none(image)
        saved_flag == True

    # Save Active Layer Position
    act_lay_pos = pdb.gimp_image_get_item_position(image, active_layer_id)
	
    # Create Bump Map & Save As Layer

    # make a copy of the character layer and add the layer to the image
    position = pdb.gimp_image_get_item_position(image, active_layer_id)
	
	
    bumpmap_layer = pdb.gimp_layer_new_from_drawable(active_layer_id, image)
    pdb.gimp_image_insert_layer(image, bumpmap_layer, parent, act_lay_pos)
    pdb.gimp_item_set_name(bumpmap_layer,"BR Logo") # rename the layer
    pdb.gimp_edit_clear(bumpmap_layer)
    pdb.gimp_selection_all(image)
    pdb.gimp_context_set_foreground(white)
    pdb.gimp_edit_bucket_fill(bumpmap_layer, 0, 0, 100, 0, 0, 0, 0)
    pdb.gimp_selection_none(image)
    pdb.gimp_item_set_visible(bumpmap_layer, True)
    bumptext_layer = pdb.gimp_layer_new_from_drawable(active_layer_id, image)
    pdb.gimp_image_insert_layer(image, bumptext_layer, parent, act_lay_pos)
	
    #  Clear Away Outside of Selection
    if saved_flag == True:
        pdb.gimp_image_select_item(image, 2, saved_selection)
        pdb.gimp_selection_invert(image) # invert selection
        pdb.gimp_edit_clear(bumptext_layer)
        pdb.gimp_selection_invert(image)
        pdb.gimp_selection_none(image)
    
    # Gausian Blur For Curves
    if blur_depth > 0:
    
        pdb.plug_in_colorify(image, bumptext_layer, white) # Color for alpha curves
		
        pdb.plug_in_gauss(image, bumptext_layer, blur_depth, blur_depth, 1) # Gaussien blur the shadow (radius blur)
	
        if contour>0: # Apply contour if necessary
            proc_apply_curves(bumptext_layer, contour)
            pdb.gimp_image_set_active_layer(image, bumptext_layer)
			
    pdb.plug_in_colorify(image, bumptext_layer, black) # Ensure Drawable BLACK for ML Bevel Reflect
    pdb.gimp_item_set_name(bumptext_layer,"Bump Text") # rename the layer
    pdb.gimp_item_set_visible(bumptext_layer, True)
	
    bumpmap_layer = pdb.gimp_image_merge_down(image, bumptext_layer, 0) # Merge Down 
    pdb.gimp_invert(bumpmap_layer) # Invert it
	

	
    pdb.gimp_image_set_active_layer(image, bumpmap_layer)
	
    #### No dialog map-type option to create just a normal map
    #if map_type == 5: # Create a Normal Map

    image_height = pdb.gimp_drawable_height(bumpmap_layer) # get height
    image_width = pdb.gimp_drawable_width(bumpmap_layer) # get width
	
					
	
    pdb.plug_in_mlbevelreflect2(image, 
                                bumpmap_layer, 
                                False, 
                                slope_options+1, 
                                flat_bevel_smoothing, 
                                flat_bevel_depth, 
                                round1_bevel_depth, 
                                round2_bevel_depth, 
                                emap_drawable, 
                                diffuse_shiny)
								
    bumpmap_layer = pdb.gimp_image_get_active_drawable(image)

								
    #  Edge Smoothing Attempt
    pdb.gimp_image_select_item(image, 0, active_layer_id)
	
    if edge_smoothing == 3:
        edge_smoothing = 2
        blur_radius = 4
		
    if edge_smoothing == 4:
        edge_smoothing = 2
        blur_radius = 5
		
		
    pdb.gimp_selection_shrink(image, edge_smoothing) # Shrink by edge smoothing amount
    pdb.gimp_selection_invert(image) # Invert selection
	
	
    if edge_smoothing > 0 and blur_radius > 0: 
        pdb.plug_in_gauss(image, bumpmap_layer, blur_radius, blur_radius, 1) # Blur Edge 
        
    
    pdb.gimp_edit_clear(bumpmap_layer) # remove unwanted colour from BRL
	
    pdb.gimp_selection_invert(image) # Invert selection
    pdb.gimp_selection_none(image) # Clear selection
								
								
								
    # identify topological layer generated by the call to the script and remove it
    br_layer = pdb.gimp_image_get_layer_by_name(image, "BR Logo")
    pdb.gimp_image_set_active_layer(image, br_layer)
	
    if desaturate_option > 0: 
        pdb.gimp_drawable_desaturate(br_layer, desaturate_option-1)
        mode = 47 # grain merge
        bumpmap_layer.mode = mode # apply mode   
        # allow user to desaturate the layer generated by the BRL plug-in before applying mode effect
    else:
        pdb.gimp_edit_clear(active_layer_id) # clear contents of original character layer
	
	
    # get the name of the topological map layer generated for each map type by BRL
    if map_type == 0:
        topo_layer = pdb.gimp_image_get_layer_by_name(image, "Spherical Topo")
    elif map_type == 1:
        topo_layer = pdb.gimp_image_get_layer_by_name(image, "Spherical Cloud")
    elif map_type == 2:
        topo_layer = pdb.gimp_image_get_layer_by_name(image, "Spherical Topo")
    elif map_type == 3:
        topo_layer = pdb.gimp_image_get_layer_by_name(image, "Spherical Cloud")
    elif map_type == 4:
        topo_layer = pdb.gimp_image_get_layer_by_name(image, (character_name + " copy"))
		
	
    
    bumpmap_layer = pdb.gimp_image_merge_down(image, bumpmap_layer, 0) # merge down the copy of the character layer
    pdb.gimp_item_set_name(bumpmap_layer,character_name) # reinstate the original name of the character
    if trim_value > 0: proc_trim_character(image, bumpmap_layer, trim_value) # allow user to trim pixels from the selection outline

    active_layer_id = pdb.gimp_image_get_active_layer(image) # identify the new active layer

    return active_layer_id
			
#############################################################################################################################

def proc_generate_emap(image, random_number, active_layer_id, emap_drawable, emap_final_size, parent, act_lay_pos, 
                    detail_level, horizontal_texture_size, vertical_texture_size, map_type, 
                    bevel_gradient, saved_fg_color, saved_bg_color, blend_direction, specular_highlight, 
                    light_direction, paint_mode, blend_shape, blur_radius):
					
    #### This is basically GnuTux's code in Python
    emap_drawable = pdb.gimp_layer_new_from_drawable(active_layer_id, image) # New Layer for Emap
    pdb.gimp_image_insert_layer(image, emap_drawable, parent, act_lay_pos + 1) # Add Layer
    pdb.gimp_image_set_active_layer(image, emap_drawable)
    pdb.gimp_layer_resize(emap_drawable, emap_final_size + 40, emap_final_size + 40, 0, 0)
    pdb.gimp_item_set_visible(emap_drawable, False)
	
    # Call Solid Noise Plug-in
    
    pdb.plug_in_solid_noise(image, emap_drawable, True, True, random_number, detail_level, horizontal_texture_size, vertical_texture_size)

    # Auto Stretch
	
    pdb.plug_in_autostretch_hsv( image, emap_drawable)
	
    # Gaussian Blur
	
    pdb.plug_in_gauss(image, emap_drawable, 5, 5, 0)

    # Set Specified Gradient Active Unless Active Gradient Is Selected
	
    if map_type < 2:
        pdb.gimp_context_set_gradient(bevel_gradient)
	
			
    # Environment Map Type: Topographical
    pdb.gimp_context_set_foreground(saved_fg_color)
    pdb.gimp_context_set_background(saved_bg_color)
	
    if map_type == 0 or map_type == 2:
        pdb.gimp_context_set_gradient(bevel_gradient)
        pdb.plug_in_gradmap(image, emap_drawable) # Map Gradient
        pdb.gimp_item_set_name(emap_drawable,"Spherical Topo") # Name emap


###############################################################################################################

	
    #  Environment Map Type: Cloud
    if paint_mode > 0 or map_type == 1 or map_type == 3:
        image_height = pdb.gimp_drawable_height(emap_drawable)
        image_width = pdb.gimp_drawable_width(emap_drawable)
		
    # Calculate Gradient Blend Start & End Points
	
        if blend_direction ==0: # Top-Bottom 
            xblend_start =  0
            yblend_start =  0
            xblend_end =  0
            yblend_end =  image_height
			
        elif blend_direction == 1: # Bottom-Top
            xblend_start =  0
            yblend_start =  image_height
            xblend_end =  0
            yblend_end =  0
			
        elif blend_direction == 2: # Left-Right
            xblend_start =  0
            yblend_start =  0
            xblend_end =  image_width
            yblend_end =  0
			
        elif blend_direction == 3: # Right-Left
            xblend_start =  image_width
            yblend_start =  0
            xblend_end = 0
            yblend_end =  0
	
        elif blend_direction == 4: # Diag-Top-Left
            xblend_start =  0
            yblend_start =  0
            xblend_end =  image_width
            yblend_end =  image_height
			
        elif blend_direction == 5: # Diag-Top-Right
            xblend_start =  image_width
            yblend_start =  0
            xblend_end =  0
            yblend_end =  image_height
			
        elif blend_direction == 6: # Diag-Bottom-Left
            xblend_start =  0
            yblend_start =  image_height
            xblend_end =  image_width
            yblend_end =  0
			
        elif blend_direction == 7: # Diag-Bottom-Right
            xblend_start =  image_width
            yblend_start =  image_height
            xblend_end =  0
            yblend_end =  0
			

    # Check Paint Mode
    if paint_mode == 0 and (map_type == 1 or map_type == 3):
        paint_mode = 4
        
        pdb.gimp_edit_blend(emap_drawable, BLEND_CUSTOM, paint_mode, blend_shape, 100, 0, 0,False, True, 3, 0.20, True, xblend_start, yblend_start, xblend_end, yblend_end)
			
	
    if map_type == 1 or map_type == 3:
        pdb.gimp_item_set_name(emap_drawable,"Spherical Cloud") # Name emap
					
######################################################################################################
    
    
    # Create Solid Color Spherical Map Using Current Foreground Color

			
    if map_type == 4:
	
        pdb.gimp_context_set_foreground(saved_fg_color)
        pdb.gimp_layer_add_alpha(emap_drawable)
        pdb.gimp_edit_clear(emap_drawable)
        pdb.gimp_drawable_fill(emap_drawable, 0) # fg fill
		
    # Gaussian Blur Final eMap
	
    if blur_radius>0:
	
        pdb.plug_in_gauss(image, emap_drawable, blur_radius, blur_radius, 0)
		
    # Make Environment Map Spherical
    if not map_type == 4:
        pdb.gimp_layer_add_alpha(emap_drawable)


    # replaced -1 values with None for python foo

    pdb.plug_in_map_object(image, emap_drawable, 1, 
                            0.5, 0.5, 2.0, 
                            0.5, 0.5, 0.0, 
                            1.0, 0.0, 0.0, 
                            0.0, 1.0, 0.0, 
                            0.0, 0.0, 0.0, 
                            0, white, 
                            -0.5, -0.5, 2.0, 
                            -1.0, -1.0, 1.0 , 
                            0.3, 1.0, 0.45, 
                            specular_highlight, 30.0, 
                            True, True, False, True, 
                            .44, 0.5, 0.5, 
                            0.5, 1.00, 
                            None, None, None, None, None, None, None, None)

    pdb.plug_in_autocrop_layer(image, emap_drawable)
	
	
    # Rotate Spherical Map For Lighting Direction 
	
    pdb.gimp_item_transform_rotate(emap_drawable,(((light_direction + 40) *(math.pi))) / 180, True, 0, 0)
	
    pdb.gimp_layer_set_offsets(emap_drawable, 0, 0)
	
    # End of emap render stuff - so return emap layer for ML BRL to use
	
    return emap_drawable
			
	
#### End of proc generate_emap

######################################################################################################################################

def fn_test_for_custom_layer(img,l_id,l_name):
    # checks that the current layer is compatible with custom fonts
    test = False
    if len(l_name) == 1:
        test = True
    elif len(l_name) == 5 or len(l_name) == 6:
        
        if ".png" in l_name:
            test = True
            if len(l_name) == 6:
                part_name_val = unicode(l_name[0:2])
                if part_name_val <= 65535: test = False
    # select the contents of the layer
    pdb.gimp_image_select_item(img,2,l_id)
    # Test to see if the selection actually contains something
    sel_is_empty = pdb.gimp_selection_is_empty(img)
    # If there are no pixels to work with flag the layer as not viable to work on
    if sel_is_empty: test = False
    # Deselect
    pdb.gimp_selection_none(img)
    return test
	
######################################################################################################################################

    #### Useful Functions
# -------------------------------------------------------------------------------------
	
def fn_selection_info(im):
    # returns selection info
    selection,x1,y1,x2,y2 = pdb.gimp_selection_bounds(im)
    s_width, s_height = x2 - x1, y2 - y1
	
    return selection, x1, x2, y1, y2, s_width, s_height


######################################################################################################################################
	
def fn_selection_amount(sel_value,r_ness,adjustment):
    # 
    value = (math.sqrt(sel_value)*r_ness)*adjustment
    return value

######################################################################################################################################
	
def proc_messagebox(message):
    old_handler = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2)
    pdb.gimp_message(str(message))
    pdb.gimp_message_set_handler(old_handler)

######################################################################################################################################
	
def proc_trim_character(img, layer, amount):    
    # select the contents of the layer
    pdb.gimp_image_select_item(img, 2, layer)
    # shrink the selection
    pdb.gimp_selection_shrink(img, amount)
    # invert selection
    pdb.gimp_selection_invert(img)
    # trim away pixels
    successful = pdb.gimp_edit_cut(layer)
    # deselect
    pdb.gimp_selection_none(img)
	
######################################################################################################################################
		
def proc_trim_away(img, layer, amount):
    #### Trim off pixels from a layer
    # set our feather settings
    our_feather = True
    our_feather_radius = 2
    #get feather settings
    feather = pdb.gimp_context_get_feather()
    feather_x, feather_y = pdb.gimp_context_get_feather_radius()
    # make sure that feather is set
    if not feather: pdb.gimp_context_set_feather(our_feather)
    # use our feather radius
    pdb.gimp_context_set_feather_radius(our_feather_radius, our_feather_radius)
    # select the contents of the layer
    pdb.gimp_image_select_item(img, 2, layer)
    # shrink the selection
    pdb.gimp_selection_shrink(img, amount)
    # invert selection
    pdb.gimp_selection_invert(img)
    # trim away pixels
    successful = pdb.gimp_edit_cut(layer)
    # deselect
    pdb.gimp_selection_none(img)
    # reset feather settings
    pdb.gimp_context_set_feather(feather)
    pdb.gimp_context_set_feather_radius(feather_x, feather_y)
	
    #### End of Proc_trim_away
	
######################################################################################################################################
	
register(
    "python_fu_Custom_Font_Bevel_Reflect_BRL_1_6_V_1_03",
    "Applies ML Bevel Reflect 2 plug-in to custom font file characters",
    "Applies ML Bevel Reflect 2 plug-in to custom font file characters",
    "Original code by GnuTux Python version by Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2013 2020",
    "<Image>/Filters/Custom Font Tools/Custom Font Bevel Reflect 1.6 V 1.03...",
    "RGB*",
    [
    (PF_SPINNER, "pf_shine_char", "Trim character in pixels: ", 0,(0,10,1)),
    (PF_OPTION, "desaturate_option", "Desaturate emap result options\n(for applying to existing faces): ", 0, ("None applied", "Desaturate Lightness", "Desaturate Luma", "Desaturate Average", "Desaturate Luminance", "Desaturate Value")),
    (PF_OPTION, "Slope_option", "Slope Options: ", 0, ("Flat Bevel", "Round1 Bevel", "Round2 Bevel")),
    (PF_SPINNER, "Flat_bevel_smoothing", "Flat Bevel - Smoothing: ", 8.0, (1.0, 32.0, 1.0)),
    (PF_SPINNER, "Flat_bevel_depth", "Flat Bevel - Depth: ", 10.0, (1.0, 10.5, 0.5)),
    (PF_SPINNER, "Round1_depth", "Round1 Bevel - Depth: ", 7.0, (1.0, 8.0, 0.5)),
    (PF_SPINNER, "Round2_depth", "Round2 Bevel - Depth: ", 20.0, (1.0, 25.0, 0.5)),
    (PF_SPINNER, "Diffuse_Shiny", "(Diffuse) 0<--->1 (Shiny): ", 1.0, (0.0, 1.0, 0.1)),
    (PF_OPTION, "Contour_option", "Contour Options: ", 0, ("Linear", "Outline", "Outline (Inner Bevel)", "Outline (Outer Bevel)", "Step Bevel", "Ring (Narrow)", "Ring (Medium)", "Ring (Wide)", "Ring (Extra Wide)", "Ring (Thick)", "Double Ring", "Double Ring (Narrow)", "Double Ring (Wide)")),
    (PF_SPINNER, "Contour_depth", "Contour Depth: ", 0, (0, 20, 1)),
    (PF_SPINNER, "Edge_smoothing", "Edge Smoothing Amount: ", 1, (0, 4, 1)),
    (PF_FILENAME, "pf_emap_name", "Spherical Emap (png/xcf): ", "none"),
    (PF_OPTION, "Environment_Map_Type", "Environment Map Type: ", 0, ("Spherical Topographical", "Spherical Cloud", "Spherical Topo (Active Gradient)", "Spherical Cloud (Active Gradient)", "Spherical Solid (Active Color)")),
    (PF_OPTION, "Blending_Mode", "Blending Mode: ", 0, ("Normal", "Dissolve", "Behind", "Multiply", "Screen", "Overlay", "Difference", "Addition", "Subtract", "Darken", "Lighten", "Hue", "Saturation", "Color", "Value", "Divide", "Dodge", "Burn", "Hard Light", "Soft Light", "Grain Extract", "Grain Merge")),
    (PF_OPTION, "Blend_Direction", "Blend Direction: ", 0, ("Top-Bottom", "Bottom-Top", "Left-Right", "Right-Left", "Diag-Top-Left", "Diag-Top-Right", "Diag-Bottom-Left", "Diag-Bottom-Right")),
    (PF_OPTION, "Blend_Shape", "Blend Shape: ", 0, ("Linear", "Bilinear", "Radial", "Square", "Conical-Symmetric", "Conical-Asymmetric", "Shapeburst-Angular", "Shapburst-Spherical", "Shapeburst-Dimpled", "Spiral-CW", "Spiral-CCW")),
    (PF_SPINNER, "Light_Direction", "Light Direction (Degrees): ", 315, (0, 360, 45)),
    (PF_SPINNER, "Emap_Specular_Highlights", "Emap Specular Highlights: ", 0.35, (0.10, 0.65, 0.05)),
    (PF_SPINNER, "Detail_Level", "Detail Level: ", 0, (0, 3, 1)),
    (PF_SPINNER, "Horizontal_Texture_Size", "Horizontal Texture Size: ", 1.2, (1.0, 5.0, 0.1)),
    (PF_SPINNER, "Vertical_Texture_Size", "Vertical Texture Size: ", 2.6, (1.0, 5.0, 0.1)),
    (PF_SPINNER, "EMap_Final_Blur", "EMap Final Blur: ", 0, (0, 20, 1)),
    (PF_OPTION, "Char_option", "Emap to character Options: ", 0, ("Single emap", "Random emap", "Character grouping emaps", "Vowel/consonant grouping emaps")),
    (PF_TOGGLE, "pf_save_emap", "Save any generated emaps\nin a new image: ", 0),
    ],
    [],
    do_stuff)

main()
